//
//  main.cpp
//  OpenNI+NITE+OpenCV
//
//  Created by Rishad Bharucha on 4/30/15.
//  Copyright (c) 2015 Rishad Bharucha. All rights reserved.
//

#include "stdafx.h"
#ifdef _WIN32
#define isnan _isnan
#endif

// STL Header
#include <iostream>

// OpenCV Header
#include <opencv2/core/core.hpp>
#include <opencv2/highgui/highgui.hpp>
#include <opencv2/imgproc/imgproc.hpp>

// o1. OpenNI Header
#include "OpenNI.h"

// n1. NiTE Header
#include "NiTE.h"

#include "kinectbvh.h"

#define MAX_KINECT_NUM 4
#define MAX_CALIBRATE_TIME_IN_SECOND 10
#define MIRROR_JOINTS false

// namespace
using namespace std;
using namespace openni;
using namespace nite;

bool m_bRecording = false;
bool m_bIsCalibrated = false;
KinectBVH *m_pKinectBVH = NULL;

Mat4 CalibrateKinect(
                     const Vec3& leftTorsoPos,
                     const Vec3& rightTorsoPos,
                     const Vec3& leftFootPos,
                     const Vec3& rightFootPos,
                     const Vec3& headPos) {
    // calculate feet position
    Vec3 feetPos = vec3_mul_scalar(vec3_add(leftFootPos, rightFootPos), 0.5f);
    // calculate three axises
    Vec3 axis_x = vec3_sub(rightTorsoPos, leftTorsoPos);
    Vec3 axis_y = vec3_sub(headPos, feetPos);
    Vec3 axis_z = vec3_zero;
    
    // calculate inverse rotation
    Mat3 inverse_rotation = mat3_inverse(mat3_from_axis(axis_x, axis_y, axis_z));
    
    // calculate inverse translation
    Vec3 inverse_translation = vec3_mul_scalar(feetPos, -1.0f);
    
    // calculate inverse scale, we assume the standard length
    // from left torso to right torso is 0.2 meters.
    float inverse_scale = 200.0f / vec3_length(axis_x);
    
    // calculate calibrate matrix
    Mat4 ret = mat4_identity;
    // move the body 1.8 meters away from the shared camera
    ret = mat4_multiply(mat4_translatef(0.0f, 0.0f, -1800.0f), ret);
    // inverse scale
    ret = mat4_multiply(mat4_scale(vec3_create(inverse_scale, inverse_scale, inverse_scale)), ret);
    // inverse rotate
    ret = mat4_multiply(mat4_from_mat3(inverse_rotation), ret);
    // inverse translate
    ret = mat4_multiply(mat4_translate(inverse_translation), ret);
    return ret;
}

inline bool IsRecording() {
    return m_bRecording;
}

inline bool IsCalibrated() {
    return m_bIsCalibrated;
}

void StartRecording()
{
    m_pKinectBVH = new KinectBVH();
    m_bRecording = true;
}

void StopRecording()
{
    m_bRecording = false;
    time_t nowtime = time(NULL);
    struct tm *local = localtime(&nowtime);
    char buf[256];
    sprintf(
            buf,
#ifdef _WIN32
            "..\\data\\%d-%d-%d-%d-%d-%d.bvh",
#elif __APPLE__
            "../data/%d-%d-%d-%d-%d-%d.bvh",
#endif
            local->tm_year+1900,
            local->tm_mon+1,
            local->tm_mday,
            local->tm_hour,
            local->tm_min,
            local->tm_sec);
    m_pKinectBVH->SaveToBVHFile(buf, MIRROR_JOINTS);
    delete m_pKinectBVH;
    m_pKinectBVH = NULL;
    m_bIsCalibrated = false;
}

void CalibrateSkeleton()
{
    m_pKinectBVH->CalibrateSkeleton();
    m_bIsCalibrated = true;
}

void drainBuffers(VideoStream** pStreams, int streamCount, VideoFrameRef* pFrame)
{
    while (true) {
        int changedIndex;
        openni::Status rc = openni::OpenNI::waitForAnyStream(
                                                             pStreams,
                                                             streamCount,
                                                             &changedIndex,
                                                             0);
        
        if (rc != openni::STATUS_OK) {
            break;
        }
        
        pStreams[changedIndex]->readFrame(pFrame);
        printf("Draining buffers ...\n");
    }
}

// sRGB: Y709 = 0.21R + 0.72G + 0.07B
inline float getBrightness(cv::Mat& color_mat, const int x, const int y)
{
    uint8_t* color_buffer = color_mat.ptr();
    int index = (y * 640 + x) * 3;
    uint8_t b = color_buffer[index];
    uint8_t g = color_buffer[index+1];
    uint8_t r = color_buffer[index+2];
    return 0.21f * r + 0.72f * g + 0.07f * b;
}

int main( int argc, char **argv )
{
    // o2. Initial OpenNI
    OpenNI::initialize();
    
    // o3. Open Device
    openni::Array<DeviceInfo> deviceInfoList;
    OpenNI::enumerateDevices(&deviceInfoList);
    int deviceCount = deviceInfoList.getSize();
    
    // no active device
    if (deviceCount == 0) {
        OpenNI::shutdown();
        return -1;
    }
    
    // limit depth sensors count
    if (deviceCount > MAX_KINECT_NUM) {
        deviceCount = MAX_KINECT_NUM;
    }
    
    Device  mDevice[MAX_KINECT_NUM];
    for (int i=0; i<deviceCount; i++) {
        mDevice[i].open( deviceInfoList[i].getUri() );
    }
    // o4. create depth stream
    VideoStream mDepthStream[MAX_KINECT_NUM];
    for (int i=0; i<deviceCount; i++) {
        mDepthStream[i].create( mDevice[i], SENSOR_DEPTH );
    }
    // o4a. set video mode
    VideoMode mDMode;
    mDMode.setResolution( 640, 480 );
    mDMode.setFps( 30 );
    mDMode.setPixelFormat( PIXEL_FORMAT_DEPTH_1_MM );
    for (int i=0; i<deviceCount; i++) {
        mDepthStream[i].setVideoMode( mDMode);
    }
    
    // o5. Create color stream
    VideoStream mColorStream[MAX_KINECT_NUM];
    for (int i=0; i<deviceCount; i++) {
        mColorStream[i].create( mDevice[i], SENSOR_COLOR );
    }
    // o5a. set video mode
    VideoMode mCMode;
    mCMode.setResolution( 640, 480 );
    mCMode.setFps( 30 );
    mCMode.setPixelFormat( PIXEL_FORMAT_RGB888 );
    for (int i=0; i<deviceCount; i++) {
        mColorStream[i].setVideoMode( mCMode);
    }
    
    // o6. image registration
    for (int i=0; i<deviceCount; i++) {
        mDevice[i].setImageRegistrationMode( IMAGE_REGISTRATION_DEPTH_TO_COLOR );
    }
    
    // n2. Initial NiTE
    NiTE::initialize();
    
    // n3. create user tracker
    UserTracker mUserTracker[MAX_KINECT_NUM];
    for (int i=0; i<deviceCount; i++) {
        mUserTracker[i].create( &mDevice[i] );
        mUserTracker[i].setSkeletonSmoothingFactor( 0.0f );
    }
    
    // create OpenCV Window
    for (int i=0; i<deviceCount; i++) {
        // get device number
        char buffer[16];
        sprintf(buffer, "%d", i);
        cv::namedWindow( "User Image " + string(buffer),  CV_WINDOW_AUTOSIZE );
    }
    
    // p1. start
    for (int i=0; i<deviceCount; i++) {
        mColorStream[i].start();
        mDepthStream[i].start();
    }
    
    // Drain buffers to get rid of asynchronized data
    VideoStream* pStreams[2*MAX_KINECT_NUM];
    VideoFrameRef dummyFrameRef;
    for (int i=0; i<deviceCount; i++) {
        pStreams[2*i] = &mColorStream[i];
        pStreams[2*i+1] = &mDepthStream[i];
    }
    drainBuffers(pStreams, 2*deviceCount, &dummyFrameRef);
    
    bool calibrated = false;
    Mat4 calibration_mat[MAX_KINECT_NUM];
    bool track_status[MAX_KINECT_NUM];
    float flip_status[MAX_KINECT_NUM];
    for (int i=0; i<deviceCount; i++) {
        track_status[i] = false;
        flip_status[i] = 0.0f;
    }
    
    int all_tracking_counter = 0;
    while( true )
    {
        // main loop
        // p2 - p5 ...
        
        // p2. prepare background
        cv::Mat cImageBGR[MAX_KINECT_NUM];
        // p3. get user frame
        VideoFrameRef mColorFrame[MAX_KINECT_NUM];
        for (int i=0; i<deviceCount; i++) {
            // p2a. get color frame
            mColorStream[i].readFrame( &mColorFrame[i] );
            // p2b. convert data to OpenCV format
            const cv::Mat mImageRGB = cv::Mat( mColorFrame[i].getHeight(),
                                              mColorFrame[i].getWidth(),
                                              CV_8UC3, (void*)mColorFrame[i].getData() );
            // smooth the image
            cv::blur(mImageRGB, mImageRGB, cv::Size(3, 3));
            // p2c. convert form RGB to BGR
            cv::cvtColor( mImageRGB, cImageBGR[i], CV_RGB2BGR );
        }
        
        // p3. get user frame
        UserTrackerFrameRef  mUserFrame[MAX_KINECT_NUM];
        for (int i=0; i<deviceCount; i++) {
            mUserTracker[i].readFrame( &mUserFrame[i] );
        }
        
        vector<Joint> merged_joints(JOINT_SIZE);
        vector<Joint> raw_joints[MAX_KINECT_NUM];
        for (int i=0; i<deviceCount; i++) {
            raw_joints[i].resize(JOINT_SIZE);
        }
        // p4. get users data
        for (int i=0; i<deviceCount; i++) {
            const nite::Array<UserData>& aUsers = mUserFrame[i].getUsers();
            // at least one user
            if (aUsers.getSize()) {
                // we only track the first user
                const UserData& rUser = aUsers[0];
                
                // p4a. check user status
                if( rUser.isNew() )
                {
                    // start tracking for new user
                    mUserTracker[i].startSkeletonTracking( rUser.getId() );
                }
                
                if( rUser.isVisible() )
                {
                    // p4b. get user skeleton
                    const Skeleton& rSkeleton = rUser.getSkeleton();
                    if( rSkeleton.getState() == SKELETON_TRACKED )
                    {
                        // Fill joints
                        for (int j = 0; j < JOINT_SIZE; j++)
                        {
                            // fill joints
                            nite::Point3f pos = rSkeleton.getJoint((nite::JointType)j).getPosition();
                            raw_joints[i][j].pos.x = pos.x;
                            raw_joints[i][j].pos.y = pos.y;
                            // convert to right hand coordinate
#ifdef _WIN32
                            raw_joints[i][j].pos.z = (MIRROR_JOINTS ? -pos.z : pos.z);
#elif __APPLE__
                            raw_joints[i][j].pos.z = (MIRROR_JOINTS ? pos.z : -pos.z);
#endif
                            raw_joints[i][j].tracked =
                            (rSkeleton.getJoint((nite::JointType)j).getPositionConfidence() > 0.5f);
                        }
                        
                        // p4c. build joints array
                        SkeletonJoint aJoints[15];
                        aJoints[ 0] = rSkeleton.getJoint( nite::JOINT_HEAD );
                        aJoints[ 1] = rSkeleton.getJoint( nite::JOINT_NECK );
                        aJoints[ 2] = rSkeleton.getJoint( nite::JOINT_LEFT_SHOULDER );
                        aJoints[ 3] = rSkeleton.getJoint( nite::JOINT_RIGHT_SHOULDER );
                        aJoints[ 4] = rSkeleton.getJoint( nite::JOINT_LEFT_ELBOW );
                        aJoints[ 5] = rSkeleton.getJoint( nite::JOINT_RIGHT_ELBOW );
                        aJoints[ 6] = rSkeleton.getJoint( nite::JOINT_LEFT_HAND );
                        aJoints[ 7] = rSkeleton.getJoint( nite::JOINT_RIGHT_HAND );
                        aJoints[ 8] = rSkeleton.getJoint( nite::JOINT_TORSO );
                        aJoints[ 9] = rSkeleton.getJoint( nite::JOINT_LEFT_HIP );
                        aJoints[10] = rSkeleton.getJoint( nite::JOINT_RIGHT_HIP );
                        aJoints[11] = rSkeleton.getJoint( nite::JOINT_LEFT_KNEE );
                        aJoints[12] = rSkeleton.getJoint( nite::JOINT_RIGHT_KNEE );
                        aJoints[13] = rSkeleton.getJoint( nite::JOINT_LEFT_FOOT );
                        aJoints[14] = rSkeleton.getJoint( nite::JOINT_RIGHT_FOOT );
                        
                        // p4d. convert joint position to image
                        cv::Point2f aPoint[15];
                        for( int  s = 0; s < 15; ++ s )
                        {
                            const Point3f& rPos = aJoints[s].getPosition();
                            mUserTracker[i].convertJointCoordinatesToDepth(
                                                                           rPos.x, rPos.y, rPos.z,
                                                                           &(aPoint[s].x), &(aPoint[s].y) );
                        }
                        
                        // we use color difference as the default flip status.
                        if (raw_joints[i][nite::JOINT_LEFT_KNEE].tracked &&
                            raw_joints[i][nite::JOINT_RIGHT_KNEE].tracked) {
                            int left_knee_x = (int)aPoint[nite::JOINT_LEFT_KNEE].x;
                            int left_knee_y = (int)aPoint[nite::JOINT_LEFT_KNEE].y;
                            int right_knee_x = (int)aPoint[nite::JOINT_RIGHT_KNEE].x;
                            int right_knee_y = (int)aPoint[nite::JOINT_RIGHT_KNEE].y;
                            // valid the coordinate
                            if (left_knee_x >= 0 && left_knee_x < 640 &&
                                left_knee_y >= 0 && left_knee_y < 480 &&
                                right_knee_x >= 0 && right_knee_x < 640 &&
                                right_knee_y >= 0 && right_knee_y < 480) {
                                // calculate brightness
                                float left_knee_brightness = getBrightness(cImageBGR[i], left_knee_x, left_knee_y);
                                float right_knee_brightness = getBrightness(cImageBGR[i], right_knee_x, right_knee_y);
                                // we assume that the left knee is always brighter than the right knee:)
                                if (left_knee_brightness < right_knee_brightness) {
                                    flip_status[i] += 0.1f;
                                    if (flip_status[i] > 1.0f) {
                                        flip_status[i] = 1.0f;
                                    }
                                } else {
                                    flip_status[i] -= 0.1f;
                                    if (flip_status[i] < -1.0f) {
                                        flip_status[i] = -1.0f;
                                    }
                                }
                                // draw circles
                                if (left_knee_x >= 20 && left_knee_x < 640-20 &&
                                    left_knee_y >= 20 && left_knee_y < 480-20) {
                                    cv::circle( cImageBGR[i], cv::Point(left_knee_x, left_knee_y), 15, cv::Scalar( 255, 255, 255 ), 2 );
                                }
                                if (right_knee_x >= 20 && right_knee_x < 640-20 &&
                                    right_knee_y >= 20 && right_knee_y < 480-20) {
                                    cv::circle( cImageBGR[i], cv::Point(right_knee_x, right_knee_y), 15, cv::Scalar( 255, 255, 255 ), 2 );
                                }
                            }
                        }
                        
                        // swap positions
                        if (!MIRROR_JOINTS) {
                            Joint temp_joint;
                            temp_joint = raw_joints[i][2];
                            raw_joints[i][2] = raw_joints[i][3];
                            raw_joints[i][3] = temp_joint;
                            temp_joint = raw_joints[i][4];
                            raw_joints[i][4] = raw_joints[i][5];
                            raw_joints[i][5] = temp_joint;
                            temp_joint = raw_joints[i][6];
                            raw_joints[i][6] = raw_joints[i][7];
                            raw_joints[i][7] = temp_joint;
                            temp_joint = raw_joints[i][9];
                            raw_joints[i][9] = raw_joints[i][10];
                            raw_joints[i][10] = temp_joint;
                            temp_joint = raw_joints[i][11];
                            raw_joints[i][11] = raw_joints[i][12];
                            raw_joints[i][12] = temp_joint;
                            temp_joint = raw_joints[i][13];
                            raw_joints[i][13] = raw_joints[i][14];
                            raw_joints[i][14] = temp_joint;
                        }
                        
                        // swap flipped positions
                        if (flip_status[i] > 0.0f) {
                            Joint temp_joint;
                            temp_joint = raw_joints[i][2];
                            raw_joints[i][2] = raw_joints[i][3];
                            raw_joints[i][3] = temp_joint;
                            temp_joint = raw_joints[i][4];
                            raw_joints[i][4] = raw_joints[i][5];
                            raw_joints[i][5] = temp_joint;
                            temp_joint = raw_joints[i][6];
                            raw_joints[i][6] = raw_joints[i][7];
                            raw_joints[i][7] = temp_joint;
                            temp_joint = raw_joints[i][9];
                            raw_joints[i][9] = raw_joints[i][10];
                            raw_joints[i][10] = temp_joint;
                            temp_joint = raw_joints[i][11];
                            raw_joints[i][11] = raw_joints[i][12];
                            raw_joints[i][12] = temp_joint;
                            temp_joint = raw_joints[i][13];
                            raw_joints[i][13] = raw_joints[i][14];
                            raw_joints[i][14] = temp_joint;
                        }
                        
                        // p4e. draw line
                        cv::line( cImageBGR[i], aPoint[ 0], aPoint[ 1], cv::Scalar( 255, 0, 0 ), 3 );
                        cv::line( cImageBGR[i], aPoint[ 1], aPoint[ 2], cv::Scalar( 255, 0, 0 ), 3 );
                        cv::line( cImageBGR[i], aPoint[ 1], aPoint[ 3], cv::Scalar( 255, 0, 0 ), 3 );
                        cv::line( cImageBGR[i], aPoint[ 2], aPoint[ 4], cv::Scalar( 255, 0, 0 ), 3 );
                        cv::line( cImageBGR[i], aPoint[ 3], aPoint[ 5], cv::Scalar( 255, 0, 0 ), 3 );
                        cv::line( cImageBGR[i], aPoint[ 4], aPoint[ 6], cv::Scalar( 255, 0, 0 ), 3 );
                        cv::line( cImageBGR[i], aPoint[ 5], aPoint[ 7], cv::Scalar( 255, 0, 0 ), 3 );
                        cv::line( cImageBGR[i], aPoint[ 1], aPoint[ 8], cv::Scalar( 255, 0, 0 ), 3 );
                        cv::line( cImageBGR[i], aPoint[ 8], aPoint[ 9], cv::Scalar( 255, 0, 0 ), 3 );
                        cv::line( cImageBGR[i], aPoint[ 8], aPoint[10], cv::Scalar( 255, 0, 0 ), 3 );
                        cv::line( cImageBGR[i], aPoint[ 9], aPoint[11], cv::Scalar( 255, 0, 0 ), 3 );
                        cv::line( cImageBGR[i], aPoint[10], aPoint[12], cv::Scalar( 255, 0, 0 ), 3 );
                        cv::line( cImageBGR[i], aPoint[11], aPoint[13], cv::Scalar( 255, 0, 0 ), 3 );
                        cv::line( cImageBGR[i], aPoint[12], aPoint[14], cv::Scalar( 255, 0, 0 ), 3 );
                        
                        // p4f. draw joint
                        for( int  s = 0; s < 15; ++ s )
                        {
                            if( aJoints[s].getPositionConfidence() > 0.5 ) {
                                bool is_right = (flip_status[i] > 0.0f ?
                                                 (s==2 || s==4 || s==6 || s==9 || s==11 || s==13) :
                                                 (s==3 || s==5 || s==7 || s==10 || s==12 || s==14));
                                cv::circle( cImageBGR[i], aPoint[s], 3, cv::Scalar( 0, is_right ? 255 : 0, 255 ), 2 );
                            }
                            else
                                cv::circle( cImageBGR[i], aPoint[s], 3, cv::Scalar( 0, 255, 0 ), 2 );
                        }
                        track_status[i] = true;
                    }
                } else {
                    if (rUser.getSkeleton().getState() == SKELETON_TRACKED && IsRecording())
                    {
                        track_status[i] = false;
                    }
                }
            }
        }
        
        if (!calibrated) {
            // all skeletons are being tracking
            bool all_tracking = true;
            for (int i=0; i<deviceCount; i++) {
                if (!track_status[i]) {
                    all_tracking = false;
                    break;
                }
            }
            
            if (all_tracking) {
                // increase counter
                all_tracking_counter++;
                // beep every one second
                if (all_tracking_counter % 30 == 0) {
                    char d=(char)(7);
                    printf("%c\n",d);
                }
            } else {
                // reset counter
                all_tracking_counter = 0;
            }
            
            // MAX_CALIBRATE_TIME seconds after all skeletons tracked
            if (all_tracking_counter == 30 * MAX_CALIBRATE_TIME_IN_SECOND) {
                // calculate calibration matrix
                for (int i=0; i<deviceCount; i++) {
                    calibration_mat[i] = CalibrateKinect(
                                                         raw_joints[i][nite::JOINT_LEFT_HIP].pos,
                                                         raw_joints[i][nite::JOINT_RIGHT_HIP].pos,
                                                         raw_joints[i][nite::JOINT_LEFT_FOOT].pos,
                                                         raw_joints[i][nite::JOINT_RIGHT_FOOT].pos,
                                                         raw_joints[i][nite::JOINT_HEAD].pos);
                }
                // calculate average torso position
                Vec3 average_torso_position = vec3_zero;
                Vec3 torso_position[MAX_KINECT_NUM];
                for (int i=0; i<deviceCount; i++) {
                    // transform raw position to calibrated coordination.
                    Vec4 pos = vec4_from_vec3(raw_joints[i][nite::JOINT_TORSO].pos, 1.0f);
                    torso_position[i] = vec3_from_vec4(mat4_mul_vector(pos, calibration_mat[i]));
                    average_torso_position = vec3_add(average_torso_position, torso_position[i]);
                }
                average_torso_position = vec3_div_scalar(average_torso_position, deviceCount);
                // adjust calibration matrix by the average torso position
                for (int i=0; i<deviceCount; i++) {
                    Vec3 offset_position = vec3_sub(average_torso_position, torso_position[i]);
                    calibration_mat[i] = mat4_multiply(calibration_mat[i], mat4_translate(offset_position));
                }
                calibrated = true;
                printf("All Kinects calibrated!\n");
            }
        } else {
            // any skeletons are being tracking
            bool any_tracking = false;
            for (int i=0; i<deviceCount; i++) {
                if (track_status[i]) {
                    any_tracking = true;
                    break;
                }
            }
            
            if(any_tracking) {
                // start recording
                if (!IsRecording()) {
                    printf("start recording\n");
                    StartRecording();
                }
                
                // calibrate T pose
                if (!IsCalibrated()) {
                    printf("calibrate skeleton\n");
                    CalibrateSkeleton();
                }
                
                // calculate the orient device
                float max_orientation_factor = FLT_MIN;
                int max_orientation_device = -1;
                for (int i=0; i<deviceCount; i++) {
                    if (track_status[i]) {
                        float orientation_factor = (MIRROR_JOINTS ?
                                                    raw_joints[i][nite::JOINT_RIGHT_HIP].pos.x -
                                                    raw_joints[i][nite::JOINT_LEFT_HIP].pos.x :
                                                    raw_joints[i][nite::JOINT_LEFT_HIP].pos.x -
                                                    raw_joints[i][nite::JOINT_RIGHT_HIP].pos.x);
                        if (orientation_factor > max_orientation_factor) {
                            max_orientation_factor = orientation_factor;
                            max_orientation_device = i;
                        }
                    }
                }
                
                // first pass of merging joints
                for (int i=0; i<deviceCount; i++) {
                    if (track_status[i] && i == max_orientation_device) {
                        for (int j = 0; j < JOINT_SIZE; j++)
                        {
                            // transform raw position to calibrated coordination.
                            Vec4 pos = vec4_from_vec3(raw_joints[i][j].pos, 1.0f);
                            pos = mat4_mul_vector(pos, calibration_mat[i]);
                            // update joints
                            merged_joints[j].pos.x = pos.x;
                            merged_joints[j].pos.y = pos.y;
                            merged_joints[j].pos.z = pos.z;
                            merged_joints[j].tracked = raw_joints[i][j].tracked;
                        }
                    }
                }
                
                // second pass of merging joints
                for (int i=0; i<deviceCount; i++) {
                    if (track_status[i] && i != max_orientation_device) {
                        for (int j = 0; j < JOINT_SIZE; j++)
                        {
                            // replace untracked position with tracked position
                            if (!merged_joints[j].tracked && raw_joints[i][j].tracked) {
                                // transform raw position to calibrated coordination.
                                Vec4 pos = vec4_from_vec3(raw_joints[i][j].pos, 1.0f);
                                pos = mat4_mul_vector(pos, calibration_mat[i]);
                                // update joints
                                merged_joints[j].pos.x = pos.x;
                                merged_joints[j].pos.y = pos.y;
                                merged_joints[j].pos.z = pos.z;
                                merged_joints[j].tracked = raw_joints[i][j].tracked;
                            }
                        }
                    }
                }
                
                // Add the positions of all joints.
                m_pKinectBVH->AddAllJointsPosition(&merged_joints[0]);
                
                // Increase the frame number.
                m_pKinectBVH->IncrementNbFrames();
            } else {
                // stop recording
                if (IsRecording()) {
                    printf("stop recording\n");
                    StopRecording();
                }
            }
        }
        
        // p5. show image
        for (int i=0; i<deviceCount; i++) {
#ifdef __APPLE__
            cv::flip(cImageBGR[i], cImageBGR[i], 1);
#endif
            // overlay the frame index
            char buffer[16];
            sprintf(buffer, "%d", mUserFrame[i].getFrameIndex());
            string text = buffer;
            int fontFace = cv::FONT_HERSHEY_SCRIPT_SIMPLEX;
            double fontScale = 1;
            int thickness = 2;
            cv::putText(cImageBGR[i], text, cv::Point(20, 35), fontFace, fontScale, cv::Scalar(255,0,0), thickness, 8);
            // get device number
            sprintf(buffer, "%d", i);
            cv::imshow( "User Image " + string(buffer), cImageBGR[i] );
            cv::moveWindow( "User Image " + string(buffer), 640*(i%3), 480*(i/3) );
        }
        
        // p6. check keyboard
        if( cv::waitKey( 1 ) == 27 ) break;
    } // end of while
    
    if (IsRecording()) {
        printf("stop recording\n");
        StopRecording();
    }
    
    if (m_pKinectBVH) {
        delete m_pKinectBVH;
        m_pKinectBVH = NULL;
    }
    
    // p7. stop
    for (int i=0; i<deviceCount; i++) {
        mUserTracker[i].destroy();
        mColorStream[i].destroy();
        mDepthStream[i].destroy();
        mDevice[i].close();
    }
    NiTE::shutdown();
    OpenNI::shutdown();
    
    return 0;
}
